// -*- Mode:C++ -*-
//Header:
//File: astack.h
//Title: array stack template
//Author: NODA Itsuki
//Date: 1998/3/25
//EndHeader:

//ModifyHistory:
// 1998/3/25: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _astack_h_
#define _astack_h_
////////////////////////////////////////////////////////////////////////

#include "base.h"

//======================================================================
// AStack<Content>
//
#ifndef DEFAULT_ARRAY_SIZE
#define DEFAULT_ARRAY_SIZE	256
#endif
#ifndef DEFAULT_EXPAND_MAG
#define DEFAULT_EXPAND_MAG	2
#endif

template <class _Contents_>
class AStack {
   public:
      //--------------------------------------------------
      // LocalTypeDef
      typedef AStack<_Contents_> astack_t ; 
      typedef _Contents_ cont_t ;

      cont_t* table ;
      UInt _top ;
      UInt size ;
      UInt mag ;
     
      inline AStack<_Contents_>(UInt sz = DEFAULT_ARRAY_SIZE,
				UInt mg = DEFAULT_EXPAND_MAG) {
				   size = 0 ;
				   mag = mg ;
				   _top = 0 ; 
				   expand(sz) ;
      } ;

      inline bool expand() { return expand(size * mag) ; } ;

      inline bool expand(UInt newsize) {
#ifdef DEBUG0
	 cerr << "Status: expanding ArrayStack:" << size << "->" << newsize
	      << endl ;
#endif	  
	 cont_t* oldtable = table ;
	 table = new cont_t[newsize] ;
	 for(UInt i = 0 ; i < size ; i++) {
	    table[i] = oldtable[i] ;
	 } 
	 size = newsize ;
	 return true ;
      } ;

      inline astack_t* push(const cont_t& val) {
	 if(_top >= size) expand() ;
	 table[_top] = val ;
	 _top++ ;
	 return this ;
      } ;

      inline cont_t pop() {
	 if(_top <= 0) {
	    cerr << "Error: pop operation hits stack bottom." << endl ;
	    exit(1) ;
	 }
	 _top-- ;
	 return table[_top] ;
      } ;
      inline cont_t popN(UInt c) {
	 if(_top < c) {
	    cerr << "Error: pop operation hits stack bottom." << endl ;
	    exit(1) ;
	 }
	 _top -= c ;
	 return table[_top] ;
      } ;

      inline cont_t top() const {
	 if(_top <= 0) {
	    cerr << "Error: top operation hits stack bottom." << endl ;
	    exit(1) ;
	 }
	 return table[_top-1] ;
      } ;

      inline cont_t topN(UInt c) const {
	 if(_top < c) {
	    cerr << "Error: top operation hits stack bottom." << endl ;
	    exit(1) ;
	 }
	 return table[_top-c-1] ;
      } ;

      inline UInt toppos() const { return _top ; } ;
      inline UInt setpos(UInt newtop) { return _top = newtop ; } ;
     
      inline friend ostream& operator<< (ostream& o, const astack_t& st) {
	 o << "#ArrayStack[size=" << st.size << ",top=" << st._top 
	   << ",mag=" << st.mag << "]" << endl ;
	 for(UInt i = st._top ; i > 0 ; i--) {
	    o << "\t" << st.table[i-1] << endl ;
	 } 
	 o << "\t___bottom___" << endl ;
	  
      } ;

};
     
////////////////////////////////////////////////////////////////////////
#endif



