// -*- Mode:C++ -*-
//Header:
//File: player.cc
//Author: NODA Itsuki
//Date: 1998/3/31
//EndHeader:

//ModifyHistory:
// 1998/3/31: Create this file
//EndModifyHistory:

//======================================================================
// Include Files
//

#include "player.h"

//======================================================================
// Function Symbol
//
GData GS_Repeat = GSymbol("$repeat",sRepeat) ;
GData GS_DoTimes = GSymbol("$dotimes",sDoTimes) ;
GData GS_End = GSymbol("$end",sNop) ;
GData GS_Nop = GSymbol("$nop",sNop) ;

GData GS_ChaseBall = GSymbol("$chaseBall",sChaseBallBetter) ;
GData GS_ChasePass = GSymbol("$chasePass",sChasePass) ;
GData GS_TurnLook = GSymbol("$turnLook",sTurnLook) ;
GData GS_FaceToBall = GSymbol("$faceToBall",sFaceToBall) ;
GData GS_FindBall = GSymbol("$findBall",sFindBall) ;
GData GS_FindBallQuit = GSymbol("$findBallQuit",sFindBallQuit) ;
GData GS_LookAround = GSymbol("$lookAround",sLookAround) ;
GData GS_LookAroundQuit = GSymbol("$lookAroundQuit",sLookAroundQuit) ;
GData GS_WaitKickOff = GSymbol("$waitKickOff",sWaitKickOff) ;
GData GS_WaitKickOffQuit = GSymbol("$waitKickOffQuit",sWaitKickOffQuit) ;
GData GS_QuitWaitKickOff = GSymbol("$quitQaitKickOff",sQuitWaitKickOff) ;
GData GS_MovetoHome = GSymbol("$moveotoHome",sMovetoHome) ;
GData GS_GotoPos = GSymbol("$gotoPos",sGotoPos) ;
GData GS_GotoHome = GSymbol("$gotoHome",sGotoHome) ;


GData GS_Cycle = GSymbol("$cycle",sCycle) ;

GData GS_top = GSymbol("top") ;
//======================================================================
// Cells
//
bool assertz(GData c, GData nm, GData def) {
   GCell* cl = c.val.symbol->gcell ;
   if(cl == NULLPTR) cl = new GCell(c.val.symbol) ;
   cl->assertz(nm,def) ;
   return true ;
} 

GData GS_C_ChaseBall = GSymbol("cellChaseBall") ;
GData GS_C_GotoHome = GSymbol("cellGotoHome") ;
GData GS_C_FindBall = GSymbol("cellFindBall") ;
GData GS_C_ConfirmPos = GSymbol("cellConfirmPos") ;
GData GS_C_ConfirmBallPos = GSymbol("cellConfirmBallPos") ;
GData GS_C_WaitPlayOn = GSymbol("cellWaitPlayOn") ;
GData GS_C_MovetoHome = GSymbol("cellMovetoHome") ;

bool initCells() {
   //  assertz(GS_C_ChaseBall,GS_top,GS_ChaseBall) ;
   assertz(GS_C_ChaseBall,GS_top,GS_ChasePass) ;
   assertz(GS_C_GotoHome,GS_top,GS_GotoHome) ;
   assertz(GS_C_FindBall,GS_top,GS_FindBall) ;
   assertz(GS_C_ConfirmPos,GS_top,GS_LookAroundQuit) ;
   assertz(GS_C_ConfirmBallPos,GS_top,GS_FindBallQuit) ;
   assertz(GS_C_WaitPlayOn,GS_top,GS_WaitKickOffQuit) ;
   assertz(GS_C_MovetoHome,GS_top,GS_MovetoHome) ;
  
   return true ;
}

//======================================================================
// Options
//


bool VerboseP = false ;
char* ServerHost = SERVERHOST ;
Int ServerPort = SERVERPORT ;
char* TeamName = TEAMNAME ;
char* PositionName = DEFAULTPOSITION ;
Float DashPower = 70 ;
Float PredFactor = 0 ;
Float ShiftFactor = 1.0 ;

Int getOption(int argc, char** argv) {
   for(Int i = 1 ; i < argc ; i++) {
      if        (!strcmp(argv[i],"-host")) {
	 i++ ;
	 ServerHost = argv[i] ;
      } else if (!strcmp(argv[i],"-port")) {
	 i++ ;
	 ServerPort = atol(argv[i]) ;
      } else if (!strcmp(argv[i],"-teamname")) {
	 i++ ;
	 TeamName = argv[i] ;
      } else if (!strcmp(argv[i],"-position")) {
	 i++ ;
	 PositionName = argv[i] ;
      } else if (!strcmp(argv[i],"-verbose")) {
	 VerboseP = true ;
      } else if (!strcmp(argv[i],"-dash")) {
	 i++ ;
	 DashPower = atof(argv[i]) ;
      } else if (!strcmp(argv[i],"-pred")) {
	 i++ ;
	 PredFactor = atof(argv[i]) ;
      } else if (!strcmp(argv[i],"-shift")) {
	 i++ ;
	 ShiftFactor = atof(argv[i]) ;
      } else {
	 cerr << "Unknown Option: " << argv[i] << endl ;
	 exit(1) ;
      }
   }
}

//======================================================================
// Main
//

int main(int argc, char** argv) {
   getOption(argc, argv) ;

   initCells() ;

   Player player ;

   player.init(ServerHost,ServerPort,TeamName,false) ;
   MSleep(100) ;

   player.posInfo = FindPosition(PositionName) ;
   if(player.posInfo->unum == 99) {
      player.posInfo = FindPosition(player.unum) ;
   } 
   if(player.posInfo->unum == 98) {
      player.posInfo->spos = xyP(0,0) ;
   } 
   cout << "My Postion is " << player.posInfo->name << endl;
   cout << "I will kick to " ;
   player.passToN = player.posInfo->passtoN ;
   player.passTo = new (PosInfo*)[player.passToN] ;
   for(Int i = 0 ; i < player.passToN ; i++) {
      player.passTo[i] = FindPosition(player.posInfo->passto[i]) ;
      cout << player.passTo[i]->name << ", " ;
   } ;
   cout << "or dribble." << endl ;

   player.sendMove(player.posInfo->spos.x, player.posInfo->spos.y) ;
   MSleep(100) ;
     
   pthread_create(&(player.sensorProcess),NULL,sensorLoop,&player) ;
   pthread_create(&(player.motorProcess),NULL,motorLoop,&player) ;
   pthread_create(&(player.listenerProcess),NULL,listenerLoop,&player) ;
   pthread_create(&(player.watchdogProcess),NULL,watchdogLoop,&player) ;
   pthread_create(&(player.planningProcess),NULL,planningLoop,&player) ;

   SleepForever() ;
}
  
//======================================================================
// Player class
//
bool Player::init(char* host, Int port, char* team, bool goalie) {
   gettimeofday(&startTime,NULLPTR) ;
   openConnection(host,port) ;
   sendInit(team,goalie) ;
   objTable.init(side) ;
   return true ;
}

bool Player::openConnection(char* host, Int port) {
   serverhost = host ;
   serverport = port ;
   sock = init_blocking_connection(host,port) ;
   return true ;
} 
     
bool Player::sendInit(char* team, bool goalie) {
   teamname = team ;
   if(goalie) {
      initInfo = send_com_init(&sock, team, (Bool)true, (Bool)true) ;
   } else {
      initInfo = send_com_init(&sock, team, (Bool)true, (Bool)false) ;
   }
   side = initInfo.side ;
   unum = initInfo.unum ;
   playmode = initInfo.playmode ;
   return true ;
} 

bool Player::updateVInfo() {
   ctime = sensorInfo.info.see.time ;
   ctimeBarrier.signal() ;
   vtime = ctime ;
   vtimeBarrier.signal() ;
   if(estimateCurrentPos()) {
      etime = ctime ;  
      updateObjTable() ;
      /* 
	 cout << "#[Pos(" << etime << ") ";
	 cout << "(" << currentPos.pos.x << ","
	 << currentPos.pos.y << ")" ;
	 cout << "<" << currentPos.dir << ">]" << endl ;
	 if(objTable.ball.etime == etime) {
	 cout << "#[Ball:(" << objTable.ball.apos.x ;
	 cout << "," << objTable.ball.apos.y << ")]" << endl ;
	 }
	 */
   }
} ;

bool Player::updateAInfo() {
   ctime = sensorInfo.info.hear.time ;
   ctimeBarrier.signal() ;
   atime = ctime ;
   atimeBarrier.signal() ;
   messageBuffer.push(&(sensorInfo.info.hear)) ;
} ;

bool Player::updateSBInfo() {
   ctime = sensorInfo.info.senseBody.time ;
   ctimeBarrier.signal() ;
   sbtime = ctime ;
   sbtimeBarrier.signal() ;
   sbInfo = sensorInfo.info.senseBody ;
} ;
     
bool Player::updateObjTable() {
   for(Int i = 0 ; i < sensorInfo.info.see.n_obj ; i++) {
      SeenObj* obj = &(sensorInfo.info.see.obj[i]) ;
      if(!obj->inViewConeP) continue ;
      switch(obj->type) {
	 case OT_Ball:
	    updateObjInfo(&(objTable.ball),obj) ; break ;
	 case OT_Goal:
	    if(obj->id.side == side)
	       updateObjInfo(&(objTable.owngoal),obj) ; 
	    else
	       updateObjInfo(&(objTable.goal),obj) ; 
	    break ;
	 case OT_Player:
	    if((obj->id.player.team == WT_Opponent) &&
	       (obj->id.player.unum != UNUM_ERR)) {
	       updateObjInfo(&(objTable.opponent[obj->id.player.unum-1]),
			     obj) ;
	    } else if ((obj->id.player.team == WT_Our) &&
		       (obj->id.player.unum != UNUM_ERR)) {
	       updateObjInfo(&(objTable.teammate[obj->id.player.unum-1]),
			     obj) ;
	    }
	    break ;
	 default:
	    break ;
      }
   }
   return true ;
} ;
		    
bool Player::updateObjInfo(ObjInfo* oinfo, SeenObj* sobj) {
   oinfo->seen.dist = sobj->dist ;
   oinfo->seen.dir = sobj->dir ;
   oinfo->seen.dist_chng = sobj->dist_chng ;
   oinfo->seen.dir_chng = sobj->dir_chng ;
   oinfo->adir = normalizeAngle(currentPos.dir + sobj->dir) ;
   oinfo->apos.x = 
      currentPos.pos.x + sobj->dist * Cos(oinfo->adir) ;
   oinfo->apos.y = 
      currentPos.pos.y + sobj->dist * Sin(oinfo->adir) ;
   oinfo->avec.x = 
      sobj->dist_chng * Cos(oinfo->adir) 
      - sobj->dist * sobj->dir_chng * DEG2RAD * Sin(oinfo->adir) ;
   oinfo->avec.y = 
      sobj->dist_chng * Sin(oinfo->adir) 
      - sobj->dist * sobj->dir_chng * DEG2RAD * Cos(oinfo->adir) ;
   oinfo->etime = ctime ;
} ;



//======================================================================
// Sensor process
//
void * sensorLoop(Player *player) {
   while(true) {
      sensorCycle(player) ;
   }
}

void sensorCycle(Player *player) {
   player->recvInfo() ;
   player->scanInfo() ;
   switch(player->sensorInfo.type) {
      case SIT_See: 
	 player->updateVInfo() ; break ;
      case SIT_Hear:
	 player->updateAInfo() ; break ;
      case SIT_SenseBody:
	 player->updateSBInfo() ; break ;
   }
}

     

//======================================================================
// Motor process
//

void * motorLoop(Player *player) {
   MSleep(1000) ;
//     player->smach.envStack.push(GS_C_ChaseBall) ;
   player->smach.envStack.push(GS_C_ChaseBall) ;
   player->smach.opStack.push(GS_Repeat, GS_Cycle, GS_top, GS_End) ;
   // player->smach.opStack.push(GS_Repeat, GS_ChaseBall, GS_End) ;
   player->smach.opStack.push(GS_WaitKickOff,GInteger(0)) ;
   player->smach.opStack.push(GS_LookAround) ; 
   while(true) {
      // trace
      //cout << player->smach.opStack << endl ;
      motorCycle(player) ;
      // Yield() ;
   }
}

void motorCycle(Player *player) {
   GData top = player->smach.opStack.pop() ;
   if(isGSymbol(top) && top.val.symbol->funcp) {
      (top.val.symbol->func)(&(player->smach)) ;
   }
}


//======================================================================
// listener process
//

void * listenerLoop(Player *player) {
   while(true) {
      listenerCycle(player) ;
   }
}

inline int strncmp1(char* str1, char* str2) {
   return strncmp(str1,str2,strlen(str1)) ;
}
inline int strncmp2(char* str1, char* str2) {
   return strncmp(str1,str2,strlen(str2)) ;
}

void listenerCycle(Player *player) {
   player->waitAInfo() ;
   UInt top = player->messageBuffer.top ;
   for(UInt i = player->listendTop ; i < top ; i++) {
      HearInfo* b = player->messageBuffer.getBuf(i) ;
      if(VerboseP) {
	 cout << "[listen:" << player->unum << ":" << b->time << "(" ;
	 switch(b->sender) {
	    case ST_Self :
	       cout << "self" ; break ;
	    case ST_Referee :
	       cout << "referee" ; break ;
	    case ST_Player:
	       cout << "player<" << b->direction << ">" ; break ;
	    default:
	       cout << "unknown" ;
	 }
	 cout << ")|" ;
	 cout << b->message << "]" << endl ;
      }
      if(b->sender == ST_Referee) {
	 if(!strncmp2(b->message,"goal_l")||
	    !strncmp2(b->message,"goal_r")||
	    !strncmp2(b->message,"half_time")||
	    !strncmp2(b->message,"time_extended")) {
	    player->wKickOffTime = b->time ;
	    player->smach.repushEnv(GS_C_WaitPlayOn) ;
	    player->smach.repushEnv(GS_C_MovetoHome) ;
	 }
      }
   }
   player->listendTop = top ;
}
//======================================================================
// watchdog process
//

void * watchdogLoop(Player *player) {
   while(true) {
      watchdogCycle(player) ;
   }
}

void watchdogCycle(Player *player) {
   player->waitVInfo() ;
   if(beforeKickOffP(player)) return ;
   //     if(VerboseP) {
   if(false) {
      player->sendSenseBody(50) ;
      cout << "[" << player->ctime << ":stamina=" 
	   << player->sbInfo.stamina ;
      cout << ", effort=" << player->sbInfo.effort ;
      cout << ", speed=" << player->sbInfo.speed << "]" << endl ;
   }
   //cout << player->smach.envStack ;
   //     cout << "(" << player->currentPos.pos.x << "," << player->currentPos.pos.y << ")" << endl ;

   if(player->ctime - player->etime > 10) {
      player->smach.repushEnv(GS_C_ConfirmPos) ;
   }
   if(player->ctime - player->objTable.ball.etime > 3) {
      player->smach.repushEnv(GS_C_ConfirmBallPos) ;
   }
   if(player->ctime - player->atime > 20) {
      player->atimeBarrier.signal() ;
   }
}	

//======================================================================
// planner process
//

void * planningLoop(Player *player) {
   while(true) {
      planningCycle(player) ;
   }
}

inline bool insideOfRectangle(Pos& pos, Pos& center, Pos& range) {
   if(fabs(pos.x - center.x) > range.x) return false ;
   if(fabs(pos.y - center.y) > range.y) return false ;
   return true ;
}



inline Pos shiftedHomePos(Player* player) {
   Pos hpos = player->posInfo->hpos ;
   Pos bpos = player->objTable.ball.apos ;
   hpos.x += bpos.x * ShiftFactor * player->posInfo->shift.x 
      / FIELDHALFLENGTH ;
   hpos.y += bpos.y * ShiftFactor * player->posInfo->shift.y
      / FIELDHALFWIDTH ;
   if(hpos.x < -(FIELDHALFLENGTH - 0.5)) 
      hpos.x = -(FIELDHALFLENGTH - 0.5) ;

   return hpos ;
}

void planningCycle(Player *player) {
   player->waitVInfo() ;
     
   ObjInfo& ball = player->objTable.ball ;

   if(beforeKickOffP(player)) return ;

   Pos shpos = shiftedHomePos(player) ;
   if((dist(ball.apos, player->currentPos.pos) < 5) || 
      insideOfRectangle(ball.apos, shpos,
			player->posInfo->range)) {
      player->smach.removeEnv(GS_C_GotoHome) ;
      player->smach.removeEnv(GS_C_FindBall) ;
   } else {
      if((dist(shiftedHomePos(player),player->currentPos.pos) < 3)&&
	 player->currentPos.pos.x > -FIELDHALFLENGTH) {
	 player->smach.removeEnv(GS_C_GotoHome) ;
	 player->smach.repushEnv(GS_C_FindBall) ;
      } else {
	 player->smach.removeEnv(GS_C_FindBall) ;
	 player->smach.repushEnv(GS_C_GotoHome) ;
      }
   }
}

//======================================================================
// Basic Operation
//

bool sNop(StackMachine* smach) { 
   return true ;
}
  
bool sRepeat(StackMachine* smach) {
   GDataStack& opStack = smach->opStack ;
   Int n = opStack.find(GS_End) ;
   if(n < 0) return false ;

   opStack.push(GS_Repeat) ;
   for(Int i = 0 ; i < n ; i++) {
      opStack.push(opStack.topN(n)) ;
   }
   return true ;
}

bool sDoTimes(StackMachine* smach) {
   GDataStack& opStack = smach->opStack ;
   GData cdt = opStack.pop() ;
   if(!isGInteger(cdt)) return false ;
   Int c = cdt.val.inum ;

   Int n = opStack.find(GS_End) ;
   if(n < 0) return false ;

   if(c <= 0) {
      opStack.popN(n) ;
      return true ;
   }

   opStack.push(GInteger(c-1)) ;
   opStack.push(GS_DoTimes) ;
   for(Int i = 0 ; i < n ; i++) {
      opStack.push(opStack.topN(n+1)) ;
   }
   return true ;
}

//======================================================================
// Cycle
//

bool sCycle(StackMachine *smach) { 
   GData mode = smach->opStack.pop() ;
   GData def ;

   def = NullSymbol ;

   smach->envLock() ;
   for(Int i = 0 ; i < smach->envStack._top ; i++) {
      GData env = smach->envStack.topN(i) ;
      if(isGSymbol(env) && env.val.symbol->gcell != NULLPTR) {
	 def = env.val.symbol->gcell->findDef(mode) ;
	 if(!NullSymP(def)) break ;
      }
   }
   smach->envUnLock() ;

   if(NullSymP(def)) {
      return false ;
   } else {
      smach->opStack.push(def) ;
      return true ;
   }
}

//======================================================================
// Basic Actions
//

bool objectExistP(Player* player) {
   Pos& ppos = player->currentPos.pos ;
   Pos& bpos = player->objTable.ball.apos ;
   Float bdist = dist(ppos,bpos) ;
   Float dd = Min(bdist,1.5) ;
   Int ctime = player->ctime ;
     
   for(Int i = 0 ; i < 11 ; i++) {
      if(ctime - player->objTable.opponent[i].etime > 5) continue ;
      Pos& opos = player->objTable.opponent[i].apos ;
      Float odist = dist(ppos,opos) ;
      if(odist < dd) {
	 Float odir = Atan2(opos.y - ppos.y,opos.x - ppos.x) ;
	 if(fabs(odist * Sin(odir)) < PLAYERSIZE * 2) 
	    return true ;
      }
      if(ctime - player->objTable.opponent[i].etime > 5) continue ;
      if(i != player->unum-1) {
	 Pos& tpos = player->objTable.teammate[i].apos ;
	 Float tdist = dist(ppos,tpos) ;
	 if(tdist < dd) {
	    Float tdir = Atan2(tpos.y - ppos.y,tpos.x - ppos.x) ;
	    if(fabs(tdist * Sin(tdir)) < PLAYERSIZE * 2) 
	       return true ;
	 }
      }
   }
   return false ;
}

bool avoidAction(Player* player) {
   if(objectExistP(player)) {
      if(VerboseP) cout << "Avoid!!" << endl ;
      player->sendTurn(90,100) ;
      player->sendDash(100,200) ;
      player->sendTurn(-90,100) ;
      return true ;
   } 
   return false ;
}

bool sChaseBallSimple(StackMachine* smach) {
   Player* player = smach->player ;
   if(player->ctime < 1) return false ;
   if((player->ctime - player->objTable.ball.etime < 10) &&
      player->ballKickable()) {
      player->sendKickAbs(100,player->objTable.goal.apos,100) ;
      player->sendTurnAbs(player->objTable.goal.apos,100) ;
      return true ;
   } 
   player->sendTurnAbs(player->objTable.ball.apos,100) ;
   player->sendDash(100,100) ;
   return true ;
}

bool sChaseBallBetter(StackMachine* smach) {
   Player* player = smach->player ;
   if(player->ctime < 1) return false ;
   if((player->ctime - player->objTable.ball.etime < 10) &&
      player->ballKickable()) {
      player->sendKickAbs(100,player->objTable.goal.apos,100) ;
      player->sendTurnAbs(player->objTable.goal.apos,100) ;
      return true ;
   } 
   Pos gpos = player->objTable.ball.apos ;
   gpos.x = gpos.x - 1 ;
   player->sendTurnAbs(gpos,100) ;
   player->sendDash(50,100) ;
   return true ;
}

inline Float absAngle(Pos& org, Pos& target) {
   return Atan2((target.y - org.y),(target.x - org.x)) ;
}

inline Float diffAngle(Pos& org, Pos& reference, Pos& target) {
   return normalizeAngle(absAngle(org,target) - absAngle(org,reference)) ;
}

Pos findGoodPassDir(Player* player) {
   PosInfo* target = NULLPTR ;
   Float dAng = 720.0 ;
   ObjInfo* ball = &(player->objTable.ball) ;
     
   for(Int i = 0 ; i < player->passToN ; i++) {
      PosInfo* t = player->passTo[i] ;
      if(t->hpos.x < player->currentPos.pos.x) continue ;
      Float dA = fabs(diffAngle(player->currentPos.pos,
				ball->apos,t->hpos)) ;
      if(t->unum == 97) dA = dA * 2.0 + 90.0 ;
      if(dAng < dA) continue ;
      target = t ;
      dAng = dA ;
   }
   if(target == NULLPTR) target = player->passTo[0] ; 
   if(VerboseP) {
      cout << "[" << player->unum << "] I pass to " << target->name 
	   << "." << endl ;
   }
   return target->hpos ;
}

bool sChasePass(StackMachine* smach) {
   Player* player = smach->player ;
   if(player->ctime < 1) return false ;
   if((player->ctime - player->objTable.ball.etime < 3) &&
      player->ballKickable()) {
      player->sendKickAbs(100,findGoodPassDir(player),100) ;
      player->sendTurnAbs(player->objTable.goal.apos,100) ;
      return true ;
   } 
   Pos gpos = player->objTable.ball.apos ;
   // predict ball pos
#ifndef NOPREDICT
   Pos gvec = player->objTable.ball.avec ;
   Float gdist = player->objTable.ball.seen.dist ;
   Float gvdir = Atan2(gvec.y,gvec.x) ;
   Float gdir = absAngle(player->currentPos.pos, gpos) ;
   Float a = fabs(Sin(gvdir - gdir) * gdist * PredFactor);
   gpos.x += a * gvec.x ;
   gpos.y += a * gvec.y ;
   if(VerboseP)
      cout << "[pred:" << a << " * (" << gvec.x << "," << gvec.y << ")]" 
	   << endl ;
#endif
   //
   gpos.x = gpos.x - 1 ;
   player->sendTurnAbs(gpos,100) ;
   avoidAction(player) ;
   player->sendDash(DashPower,100) ;
   return true ;
}     
     

bool sFaceToBall(StackMachine *smach) {
   Player* player = smach->player ;
   player->sendTurnAbs(player->objTable.ball.apos) ;
   player->waitVInfo() ;
   return true ;
}

bool sFindBall(StackMachine *smach) {
   Player* player = smach->player ;
     
   Int etime = player->objTable.ball.etime ;
   sFaceToBall(smach) ;
   while(etime >= player->objTable.ball.etime) {
      sTurnLook(smach) ;
   }
}

bool sFindBallQuit(StackMachine *smach) {
   sFindBall(smach) ;
   smach->removeEnv(GS_C_ConfirmBallPos) ;
   return true ;
}

bool sTurnLook(StackMachine* smach) {
   Player* player = smach->player ;
   player->sendTurn(90) ;
   player->waitVInfo() ;
}

bool sLookAround(StackMachine* smach) {
   Player* player = smach->player ;
   smach->opStack.push(GS_DoTimes,GInteger(4),GS_TurnLook,GS_End) ;
}

bool sLookAroundQuit(StackMachine* smach) {
   sLookAround(smach) ;
   smach->removeEnv(GS_C_ConfirmPos) ;
   return true ;
}

bool sWaitKickOff(StackMachine *smach) {
   Player* player = smach->player ;
   GData cdt = smach->opStack.pop() ;
   Int c = IntVal(cdt) ;
   if(VerboseP)
      cout << "Waiting Kick Off!! [" << player->unum << ":" 
	   << player->wKickOffTime << "==" << player->ctime << "]" << endl ;
   if(c <= 0) {
      c = 0 ;
      player->wKickOffTime = player->ctime ;
   }
   if(player->wKickOffTime < player->ctime) {
      player->wKickOffTime = -1 ;
      return true ;
   }
   smach->opStack.push(GS_TurnLook, GS_WaitKickOff, GInteger(c+1)) ;
   return false ;
}

bool sWaitKickOffQuit(StackMachine *smach) {
   smach->opStack.push(GS_WaitKickOff,GInteger(0),GS_QuitWaitKickOff) ;
   return true ;
}

bool sQuitWaitKickOff(StackMachine *smach) {
   smach->removeEnv(GS_C_WaitPlayOn) ;
   return true ;
}

bool sMovetoHome(StackMachine *smach) {
   Player* player = smach->player ;
   if(VerboseP) 
      cout << "[MovetoHome:" << player->unum << "]" << endl ;
   smach->removeEnv(GS_C_MovetoHome) ;
   MSleep(100) ;
   player->sendMove(player->posInfo->spos.x, player->posInfo->spos.y) ;
   MSleep(100) ;
   player->sendMove(player->posInfo->spos.x, player->posInfo->spos.y) ;
   MSleep(100) ;
   player->sendMove(player->posInfo->spos.x, player->posInfo->spos.y) ;
   return true ;
}
 
bool sGotoPos(StackMachine *smach) {
   Player* player = smach->player ;

   GData xdt = smach->opStack.pop() ;
   GData ydt = smach->opStack.pop() ;
   Float x = FltVal(xdt) ;
   Float y = FltVal(ydt) ;

   Pos targetPos = xyP(x,y) ;
   bool backp = false ;
   //back?
   Pos ppos = player->currentPos.pos ;
   Float ddir = diffAngle(ppos,player->objTable.ball.apos,targetPos) ;
   if(fabs(ddir) > 90.0) backp = true ;
   //

   player->sendTurnAbs(targetPos,100,backp) ;
   if(dist(player->currentPos.pos,targetPos)>10) {
      player->sendDash(20,100,backp) ;
   } else {
      player->sendDash(10,100,backp) ;
   }

   return true ;
}

bool sGotoHome(StackMachine *smach) {
   Player* player=smach->player ;

   Pos hpos = shiftedHomePos(player) ;
   GData xdt = GFloat(hpos.x) ;
   GData ydt = GFloat(hpos.y) ;

   smach->opStack.push(GS_GotoPos, xdt, ydt) ;

   return true ;
}
       
