/*
 *	Header:
 *	File: netif.c
 *	Author: Noda Itsuki
 *	Date: 1997/04/1
 *	EndHeader:
 */

#include <stdio.h>
#include <stdlib.h>

#include "libsclient.h"

const FlagInfo Flags[] = FLAGINFOTAB ;

Bool receive_info(Socket* socket, char* buffer, int size)
{
  int ret = FALSE ;
  char save = '\0';

  while(TRUE) {
    if (receive_message(socket, buffer, size)) {
      ret = TRUE ;
      save = buffer[0] ;
    }
    else
      break ;
  }

  buffer[0] = save ;
  return ret ;
}


Bool scan_info(char *buffer, SensorInfo *sinfo, char *teamname)
{
  Bool scan_visual_info(char *buffer, SeeInfo *sinfo, char *teamname) ;
  Bool scan_auditorial_info(char *buffer, HearInfo *hinfo) ;
  Bool scan_senseBody_info(char *buffer, SenseBodyInfo *sbinfo) ;
  char com[COMMAX] ;

  sscanf(buffer, "(%s ", com) ;

  if (!strcmp(com, "see")) {
    sinfo->type = SIT_See ;
    return scan_visual_info(buffer, &(sinfo->info.see), teamname) ;
  }
  else if (!strcmp(com, "hear")) {
    sinfo->type = SIT_Hear ;
    return scan_auditorial_info(buffer, &(sinfo->info.hear)) ;
  }
  else if (!strcmp(com, "sense_body")) {
       sinfo->type = SIT_SenseBody ;
       return scan_senseBody_info(buffer, &(sinfo->info.senseBody)) ;
  }
  else {
    return FALSE ;
  }

  return TRUE ;
}


static char *next_token(char *buf)
{
  int i = 0 ;

  while (TRUE) {
    if (*buf == '(')
      i++ ;
    else if (*buf == ')') {
      if (i == 0) {
	*buf = '\0' ;
	return NULL ;
      }
      i-- ;
    }
    else if (*buf == ' ' && i == 0)
      break ;
    else if (*buf == '\0')
      return NULL ;
    buf++ ;
  }

  *buf = '\0' ;
  buf++ ;

  return buf ;
}


Bool scan_visual_info(char *buffer, SeeInfo *sinfo, char *teamname)
{
  static char *next_token(char *buf) ;
  char *next, *obj, *param, *tmp ;
  int num = 0 ;

  sscanf(buffer, "(see %d ", &(sinfo->time)) ;
  buffer = next_token(buffer + 1) ;
  if(buffer != NULL) buffer = next_token(buffer) ;

  while(buffer != NULL) {
    next = next_token(buffer++) ;
    param = next_token(buffer++) ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].dist = atof(tmp);
    }
    else
      sinfo->obj[num].dist = DIST_ERR ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].dir = atof(tmp) ;
    }
    else
      sinfo->obj[num].dir = DIR_ERR ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].dist_chng = atof(tmp) ;
    }
    else
      sinfo->obj[num].dist_chng = CHNG_ERR ;

    if (param != NULL) {
      tmp = param ;
      param = next_token(tmp) ;
      sinfo->obj[num].dir_chng = atof(tmp) ;
    }
    else
      sinfo->obj[num].dir_chng = CHNG_ERR ;

    obj = next_token(buffer) ;

    if (!strcmp(buffer, "ball")) { 
      sinfo->obj[num].type = OT_Ball ;
      sinfo->obj[num].inViewConeP = TRUE ;
    }
    else if(!strcmp(buffer, "Ball")) {
      sinfo->obj[num].type = OT_Ball ;
      sinfo->obj[num].inViewConeP = FALSE ;
    }
    else if(!strcmp(buffer, "player")) {
      sinfo->obj[num].type = OT_Player ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.player.team = WT_Unknown ;
      sinfo->obj[num].id.player.unum = UNUM_ERR ;
      if (obj != NULL) {
	buffer = obj ;
	obj = next_token(buffer) ;
	if (!strcmp(buffer, teamname))
	  sinfo->obj[num].id.player.team = WT_Our ;
	else
	  sinfo->obj[num].id.player.team = WT_Opponent ;
				
	if (obj != NULL) {
	  (void)next_token(obj) ;
	  sinfo->obj[num].id.player.unum = atoi(obj) ;
	}
      }
    }
    else if(!strcmp(buffer, "Player")) {
      sinfo->obj[num].type = OT_Player ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.player.team = WT_Unknown ;
      sinfo->obj[num].id.player.unum = UNUM_ERR ;
    }
    else if(!strcmp(buffer, "goal")) {
      sinfo->obj[num].type = OT_Goal ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.side = S_UNKNOWN ;
      if (obj != NULL) {
	buffer = obj ;
	obj = next_token(buffer) ;
	if (*buffer == 'l')
	  sinfo->obj[num].id.side = S_LEFT ;
	else if (*buffer == 'r')
	  sinfo->obj[num].id.side = S_RIGHT ;
      }
    }
    else if(!strcmp(buffer, "Goal")) {
      sinfo->obj[num].type = OT_Goal ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.side = S_UNKNOWN ;
    }
    else if(!strcmp(buffer, "flag")) {
      sinfo->obj[num].type = OT_Flag ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.fpos = FP_Unknown ;
      if (obj != NULL) {
	int namelen ;
	int i ;
	buffer = obj ;
	for(namelen = 0  ; 
	    (buffer[namelen] != ')')&&(buffer[namelen] != '\000') ; 
	    namelen++) ;
	for(i = 0 ; i < FP__MAX ; i++) {
	  if(namelen == strlen(Flags[i].namestr) &&
	     !strncmp(buffer,Flags[i].namestr,namelen)) {
	    sinfo->obj[num].id.fpos = Flags[i].name ;
	    break ;
	  }
	}
      }
    }
    else if(!strcmp(buffer, "Flag")) {
      sinfo->obj[num].type = OT_Flag ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.fpos = FP_Unknown ;
    }
    else if(!strcmp(buffer, "line")) {
      sinfo->obj[num].type = OT_Line ;
      sinfo->obj[num].inViewConeP = TRUE ;
      sinfo->obj[num].id.lpos = LP_Unknown ;
      if (obj != NULL) {
	buffer = obj ;
	obj = next_token(buffer) ;
	if (*buffer == 'l')
	  sinfo->obj[num].id.lpos = LP_L ;
	else if (*buffer == 't')
	  sinfo->obj[num].id.lpos = LP_T ;
	else if (*buffer == 'b')
	  sinfo->obj[num].id.lpos = LP_B ;
	else if (*buffer == 'r')
	  sinfo->obj[num].id.lpos = LP_R ;
      }
    }
    else if(!strcmp(buffer, "Line")) {
      sinfo->obj[num].type = OT_Line ;
      sinfo->obj[num].inViewConeP = FALSE ;
      sinfo->obj[num].id.lpos = LP_Unknown ;
    }
    else {
      printf("scan_visual_info: miss obj\n") ;
      return FALSE ;
    }

    buffer = next ;
    num++ ;
  }
  sinfo->n_obj = num ;

  return TRUE ;
}


Bool scan_auditorial_info(char *buffer, HearInfo *hinfo)
{
  static char *next_token(char *buf) ;
  char tmp[64] ;
  int i ;

  sscanf(buffer, "(hear %d %s", &(hinfo->time), tmp) ;
  buffer = next_token(buffer + 1) ;
  buffer = next_token(buffer) ;
  buffer = next_token(buffer) ;

  for(i = strlen(buffer); i >= 0; i--)
    if (buffer[i] == ')') {
      buffer[i] = '\0' ;
      break ;
    }

  strcpy(hinfo->message, buffer) ;

  if (!strcmp(tmp, "referee")) {
    hinfo->sender = ST_Referee ;
    hinfo->direction = 0.0 ;
  }
  else if (!strcmp(tmp, "self")) {
    hinfo->sender = ST_Self ;
    hinfo->direction = 0.0 ;
  }
  else {
    hinfo->sender = ST_Player ;
    hinfo->direction = atof(tmp) ;
  }

  return TRUE ;
}

Bool scan_senseBody_info(char *buffer, SenseBodyInfo *sbinfo) {
     static char *next_toke(char *buf) ;
     char *next, *obj, *param, *tmp ;

     sscanf(buffer, "(see %d ", &(sbinfo->time)) ;
     buffer = next_token(buffer + 1) ;
     buffer = next_token(buffer) ;

     sbinfo->countKick = -1 ;      sbinfo->countDash = -1 ; 
     sbinfo->countTurn = -1 ;      sbinfo->countSay = -1 ; 
     sbinfo->countCatch = -1 ; 

     while(buffer != NULL) {
	  next = next_token(buffer) ;
	  buffer++ ;
	  param = next_token(buffer) ;
	  
	  if        (!strcmp(buffer,"view_mode")) {
	       char *nparam ;
	       nparam = next_token(param) ;
	       if(!strcmp(param,"high")) {
		    sbinfo->viewQuality = VQ_high ;
	       } else {
		    sbinfo->viewQuality = VQ_low ;
	       }
	       if(!strcmp(nparam,"wide)")) {
		    sbinfo->viewWidth = VW_Wide ;
	       } else if(!strcmp(nparam,"normal)")) {
		    sbinfo->viewWidth = VW_Normal ;
	       } else {
		    sbinfo->viewWidth = VW_Narrow ;
	       }
	  } else if (!strcmp(buffer,"stamina")) {
	       sscanf(param,"%lf %lf)",&sbinfo->stamina, &sbinfo->effort) ;
	  } else if (!strcmp(buffer,"speed")) {
	       sscanf(param,"%lf",&sbinfo->speed) ;
	  } else if (!strcmp(buffer,"kick")) {
	       sscanf(param,"%d",&sbinfo->countKick) ;
	  } else if (!strcmp(buffer,"dash")) {
	       sscanf(param,"%d",&sbinfo->countDash) ;
	  } else if (!strcmp(buffer,"turn")) {
	       sscanf(param,"%d",&sbinfo->countTurn) ;
	  } else if (!strcmp(buffer,"say")) {
	       sscanf(param,"%d",&sbinfo->countSay) ;
	  } else if (!strcmp(buffer,"catch")) {
	       sscanf(param,"%d",&sbinfo->countCatch) ;
	  } else {
	       printf("Illegal sense_body info:(%s %s\n",
		       buffer,param);
	       return FALSE ;
	  } 
	  buffer = next ;
     }

     return TRUE ;
}

