%%======================================================================
%% server
%%

newcell(server).
push(server).

$fact(teamname("kappa")).
$fact(shost("localhost",6000)).
%$fact(shost("rc98.soccer30",6000)).
%$fact(shost("calamus",6000)).
%$fact(shost("oak",6000)).

pop(_).
push(server,main).

%%======================================================================
%% player basic
%%

newcell(player).
push(player).

$rule(log(Desc),
	$and(verbosep,$write(Desc),$cut)).
$rule(log(_),$cut).

$rule(verbose,
	$asserta_fact(verbosep)).
$rule(silent,
	$asserta_rule(verbosep,$and($cut,$fail))).

$rule(s(Team),
	$and(	$asserta_fact(teamname(Team)),
		$cut,
		start)).
$rule(s(Team,SHost),
	$and(	$asserta_fact(teamname(Team)),
		$asserta_fact(shost(SHost,6000)),
		$cut,
		start)).


$rule(start,
	$and(	$cut,
		newplayer(P),
		$msleep(500),
		setpos(P),
		start(sensor),
		start(motor),
		$msleep(500),
		start(watchdog),
		start(listener),
		start(planner),
		pop(X))).

$rule(setpos(P),
	$and(	$s$get_unum(P,UNum),
		$write(unum(UNum)),$nl,
		findposinfo(UNum,PosName,InitPos,HomePos,Range,Shift,PassDest),
		$cut,
		$asserta_fact(unum(UNum)),
		$asserta_fact(posname(PosName)),
		$asserta_fact(initpos(InitPos)),
		$asserta_fact(homepos(HomePos,Range,Shift)),
		$asserta_fact(passtolist(PassDest)),
		$s$new_pass_dst(P,32),
		setpassdest(P,PassDest),
		$eq(InitPos,p(X,Y)),
		$s$send_move(P,X,Y))).

$rule(findposinfo(UNum,PosName,InitPos,HomePos,Range,Shift,PassDest),
	$and(	push(position_table),
		pos(UNum,PosName,InitPos,HomePos,Range,Shift,PassDest),
		pop(_),
		$cut)).
$rule(findposinfo(UNum,PosName,InitPos,HomePos,Range,Shift,PassDest),
	$and(	pop(position_table),
		$write(error_can_not_find_position_info),$nl,
		$cut,
		$fail)).

$fact(setpassdest(P,$)).
$rule(setpassdest(P,$(PosName,Rest)),
	$and(	findposinfo(UNum,PosName,InitPos,HomePos,Range,Shift,PassDest),
		$cut,
		$eq(HomePos,p(X,Y)),
		$s$add_pass_dst(P,X,Y,pass),
		setpassdest(P,Rest))).

$rule(newplayer(P),
	$and(	teamname(Teamname),
		shost(SHost,SPort),
		gencell(NewCell,player),
		push(NewCell),
		$cut,
		$s$new_player(P,Teamname,SHost,SPort),
		$asserta_fact(player(P)),
		$asserta_fact(workcell(NewCell)))).

$rule(start(Module),
	$and(	gencell(NewCell,Module),
		push(Module,NewCell),
		fork($and(pushsecond(NewCell),start),PID),
		$asserta_fact(pid(Module,PID,NewCell)))).

$rule(mainloop(Mode),
	$and($repeat,cycle(Mode),fail)).

pop(_).
push(player,main).

%%======================================================================
%% sensor
%%

newcell(sensor).
push(sensor).

$rule(start,
	$and(player(P),$s$sensor_loop(P))).

pop(_).

%%======================================================================
%% watchdog
%%

newcell(watchdog).
push(watchdog).

$rule(start,
	$and(player(P),init(P),$cut,mainloop(top(P)))).

$rule(init(P),
	$and(	pid(motor,Gem,MotorWork),
		gencell(MW_Work,mw),
		pushsecond(MW_Work,Gem),
		$asserta_fact(workcell(mw,MW_Work)),
		push(find_ball_q,MW_Work),
		push(find_ball_q,MW_Work))).
	

$rule(cycle(top(P)),
	$and($s$send_sense_body(P),fail)).
$rule(cycle(top(P)),
	$and(cycle(check_btime(P)),fail)).
$rule(cycle(top(P)),
	$and(cycle(check_etime(P)),fail)).
$rule(cycle(top(P)),
	$and($sleep(1),fail)).

$rule(cycle(top(P)),
	$and(true)).

$rule(cycle(check_btime(P)),
	$and(	$s$get_betime(P,BTime),
		$s$get_ctime(P,CTime),
		$is(D,sub(CTime,BTime)),
		$comp_gt(D,7),
		$cut,
		workcell(mw,MW),
		needs(find_ball_q,MW),
		$cut)).

$rule(cycle(check_etime(P)),
	$and(	$s$get_etime(P,ETime),
		$s$get_ctime(P,CTime),
		$is(D,sub(CTime,ETime)),
		$comp_gt(D,10),
		$cut,
		workcell(mw,MW),
		needs(look_around_q,MW),
		$cut)).

pop(_).


%%======================================================================
%% listener
%%

newcell(listener).
push(listener).

$rule(start,
	$and(player(P),init(P),$cut,mainloop(top(P)))).

$rule(init(P),
	$and(	pid(motor,Gem,MotorWork),
		gencell(ML_Work,ml),
		push(ML_Work,MotorWork),
		$asserta_fact(workcell(ml,ML_Work)))).

$rule(cycle(top(P)),
	$and(	$s$get_new_message(P,Time,Sender,Message),
		$cut,
		listen(P,Time,Sender,Message),
%		$write(hear(Time,Sender,Message)),$nl,
		$cut,fail)).
$rule(cycle(top(P)),
	$and($s$wait_ainfo(P),fail)).

$rule(cycle(top(P)),
	$true).

$rule(listen(P,Time,referee,goal(Side,Point)),
	$and(	workcell(mw,MW),
		needs(moveto_home_q,MW),
		$cut)).

$rule(listen(P,Time,referee,half_time),
	$and(	workcell(mw,MW),
		needs(moveto_home_q,MW),
		$cut)).
$rule(listen(P,Time,referee,time_extended),
	$and(	workcell(mw,MW),
		needs(moveto_home_q,MW),
		$cut)).
$rule(listen(P,Time,referee,goal_kick),
	$and(	workcell(mw,MW),
		needs(look_around_q,MW),
		$cut)).
$rule(listen(P,Time,referee,corner_kick),
	$and(	workcell(mw,MW),
		needs(look_around_q,MW),
		$cut)).
$rule(listen(P,Time,referee,kick_in),
	$and(	workcell(mw,MW),
		needs(look_around_q,MW),
		$cut)).

$rule(listen(P,Time,referee,MSG),
	$and(   $true,
%		$write(hear(Time,Sender,MSG)),$nl,
		$cut,fail)).
		
pop(_).


%%======================================================================
%% planner
%%

newcell(planner).
push(planner).

$rule(start,
	$and(player(P),init(P),$cut,mainloop(top(P)))).

$rule(init(P),
	$and(	pid(motor,Gem,MotorWork),
		gencell(MP_Work,mp),
		push(MP_Work,MotorWork),
		$asserta_fact(workcell(mp,MP_Work)))).

$rule(cycle(top(P)),
	$and($s$wait_vinfo(P),fail)).

$rule(cycle(top(P)),
	$and(cycle(check_pos(P)),fail)).
		
$rule(cycle(top(P)),
	$cut).

$rule(cycle(check_pos(P)),
	$and(	$s$kickable(P),
		$cut,
		workcell(mp,MP),
		fremove(goto_home,MP),
		fremove(find_ball,MP),
		log(immediate_chase_mode),		
		$cut)).

$rule(cycle(check_pos(P)),
	$and(	$s$get_bpos(P,BX,BY),
		homepos(p(HX,HY),p(RX,RY),p(SX,SY)),
		$s$shifted_pos(SHX,SHY,HX,HY,BX,BY,SX,SY,1.0),
%		log(chkbpos(BX,BY,SHX,SHY,RX,RY)),
		$s$inside_rectangle(BX,BY,SHX,SHY,RX,RY),
		$cut,
		workcell(mp,MP),
		fremove(goto_home,MP),
		fremove(find_ball,MP),
		log(chase_mode),		
		$cut)).
$rule(cycle(check_pos(P)),
	$and(	homepos(p(HX,HY),p(RX,RY),p(SX,SY)),
		$s$get_cpos(P,CX,CY,Dir),
		$s$get_bpos(P,BX,BY),
		$s$shifted_pos(SHX,SHY,HX,HY,BX,BY,SX,SY,1.0),
%		log(chkcpos(CX,CY,SHX,SHY,2,2)),
		$s$inside_rectangle(CX,CY,SHX,SHY,2,2),
%		log(yes),
		$cut,
		workcell(mp,MP),
		fremove(goto_home,MP),
		needs(find_ball,MP),
		log(find_mode),		
		$cut)).
$rule(cycle(check_pos(P)),
	$and(	
		workcell(mp,MP),
		fremove(find_ball,MP),
		needs(goto_home,MP),
		log(home_mode),		
		$cut)).
		

pop(_).


%%======================================================================
%% motor
%%

newcell(motor).
push(motor).

$rule(start,
	$and(	player(P),
		pid(motor,_,Work),
		push(chase_pass_ball,Work),
		mainloop(top(P)))).

$rule(cycle(turn90(P)),
	$and($cut,$s$send_turn(P,90),$s$wait_vinfo(P))).

$rule(remove_cell_work(Cell),
	$and(workcell(mw,W),$remove_a_cell(Cell,W),fail)).
$rule(remove_cell_work(Cell),
	$and(workcell(mp,W),$remove_a_cell(Cell,W),fail)).
$rule(remove_cell_work(Cell),
	$cut).

pop(_).

%%--------------------------------------------------
%% chase_ball
%%

newcell(chase_ball).
push(chase_ball).

$rule(cycle(top(P)),
	$s$chase_ball_better(P)).
$rule(cycle(top(P)),
	$and($msleep(100),$cut)).

pop(_).

%%--------------------------------------------------
%% chase_pass_ball
%%

newcell(chase_pass_ball).
push(chase_pass_ball).

$rule(cycle(top(P)),
	$s$chase_pass(P)).
$rule(cycle(top(P)),
	$and($msleep(100),$cut)).

pop(_).

%%--------------------------------------------------
%% find ball
%%

newcell(find_ball).
push(find_ball).

$rule(facetoball(P),
	$and(	$s$get_bpos(P,BX,BY),
		$s$send_turn_apos(P,BX,BY),
		$s$wait_vinfo(P),
		$cut)).

$rule(cycle(top(P)),
	$and(	$s$get_betime(P,BETime),
		facetoball(P),
		cycle(wait_new_binfo(P,BETime)),
		$cut)).

$rule(cycle(wait_new_binfo(P,BETime)),
	$and(rep,new_binfo(P,BETime),$cut)).

$rule(new_binfo(P,OldTime),
	$and(	$s$get_betime(P,NewTime),	
		$comp_gt(NewTime,OldTime),
		$cut)).
$rule(new_binfo(P,OldTime),
	$and(	cycle(turn90(P)),
		$cut,
		fail)).

$rule(rep,$true).
$rule(rep,rep).

pop(_).

%%--------------------------------------------------
%% find ball & quit
%%

newcell(find_ball_q).
push(find_ball_q).

$rule(facetoball(P),
	$and(	$s$get_bpos(P,BX,BY),
		$s$send_turn_apos(P,BX,BY),
		$s$wait_vinfo(P),
		$cut)).

$rule(cycle(top(P)),
	$and(	$s$get_betime(P,BETime),
		facetoball(P),
		cycle(wait_new_binfo(P,BETime)),
		$fail)).
$rule(cycle(top(P)),
	$and(remove_cell_work(find_ball_q),$cut)).
$rule(cycle(wait_new_binfo(P,BETime)),
	$and($repeat,new_binfo(P,BETime),$cut)).

$rule(new_binfo(P,OldTime),
	$and(	$s$get_betime(P,NewTime),	
		$comp_gt(NewTime,OldTime),
		$cut)).
$rule(new_binfo(P,OldTime),
	$and(	cycle(turn90(P)),
		$cut,
		fail)).

pop(_).

%%--------------------------------------------------
%% look around
%%

newcell(look_around_q).
push(look_around_q).

$rule(cycle(top(P)),
	$and(dotimes(I,4),cycle(turn90(P)),fail)).
$rule(cycle(top(P)),
	$and(remove_cell_work(look_around_q),$cut)).

pop(_).

%%--------------------------------------------------
%% goto home
%%

newcell(goto_home).
push(goto_home).

$rule(cycle(top(P)),
	$and(	
		homepos(p(HX,HY),p(RX,RY),p(SX,SY)),
		$s$get_bpos(P,BX,BY),
		$s$shifted_pos(X,Y,HX,HY,BX,BY,SX,SY,1.0),
		$cut,
		cycle(goto_pos(P,X,Y)))).
$rule(cycle(goto_pos(P,X,Y)),
	$and(	$s$send_turn_apos(P,X,Y),
		$msleep(100),
		$s$send_dash(P,50),
		$msleep(100),
		$cut)).
pop(_).

%%--------------------------------------------------
%% nop
%%

newcell(nop).
push(nop).

$rule(cycle(top(P)),
	$and($msleep(100),$cut)).
pop(_).



%%--------------------------------------------------
%% moveto home
%%

newcell(moveto_home_q).
push(moveto_home_q).

$rule(cycle(top(P)),
	$and(	
		initpos(p(IX,IY)),
		cycle(moveto_pos(P,IX,IY)),
		$fail)).
$rule(cycle(top(P)),
	$and(	remove_cell_work(moveto_home_q),
		workcell(mw,MW_Work),
		push(find_ball_q,MW_Work),
		push(find_ball_q,MW_Work),
		$msleep(1000),
		$cut)).
	
$rule(cycle(moveto_pos(P,X,Y)),
	$and(	$s$send_move(P,X,Y),
		$msleep(100),
		$s$send_move(P,X,Y),
		$msleep(100),
		$s$send_move(P,X,Y),
		$msleep(100),
		$cut)).
pop(_).

%%======================================================================
%% peeker_v
%%

newcell(peeker_v).
push(peeker_v).

$rule(start,
	$and(player(P),mainloop(top(P)))).

$rule(cycle(top(P)),
	$and(	$s$wait_vinfo(P),
		$s$get_rtime(P,RTime),
		$s$get_ctime(P,CTime),
		$s$get_vtime(P,VTime),
		$write(get(v,VTime,CTime,RTime)),
		$nl,$cut)).

pop(_).

%%======================================================================
%% peeker_a
%%

newcell(peeker_a).
push(peeker_a).

$rule(start,
	$and(player(P),mainloop(top(P)))).

$rule(cycle(top(P)),
	$and(	$s$wait_ainfo(P),
		$s$get_rtime(P,RTime),
		$s$get_ctime(P,CTime),
		$s$get_atime(P,VTime),
		$write(get(a,VTime,CTime,RTime)),
		$nl,$cut)).

pop(_).

%%======================================================================
%% peeker_s
%%

newcell(peeker_s).
push(peeker_s).

$rule(start,
	$and(player(P),mainloop(top(P)))).

$rule(cycle(top(P)),
	$and(	$s$wait_sinfo(P),
		$s$get_rtime(P,RTime),
		$s$get_ctime(P,CTime),
		$s$get_stime(P,VTime),
		$write(get(s,VTime,CTime,RTime)),
		$nl,$cut)).

pop(_).

%%======================================================================
%% position table
%%

newcell(position_table).
push(position_table).

%%    pos(UN, PosName,  InitPos,   HomePos,   Range,   Shift,
%%		PassDestList)

$fact(pos( 1, goalie,	p(-51,  0),p(-50,  0),p(15,15),p( 5,5),
		$(l_w_back,$(r_w_back,$(l_clear,$(r_clear,$)))))).
$fact(pos( 2, sweeper,	p(-35,  0),p(-20,  0),p(25,20),p(20,5),
		$(l_w_back,$(r_w_back,$(d_half,$(l_clear,$(r_clear,$))))))).
$fact(pos( 3, l_f_back,	p(-40, -8),p(-30, -8),p(20,30),p(20,5),
		$(l_w_back,$(l_o_half,$(l_clear,$))))).
$fact(pos( 4, r_f_back,	p(-40,  8),p(-30,  8),p(20,30),p(20,5),
		$(r_w_back,$(r_o_half,$(r_clear,$))))).
$fact(pos( 5, l_w_back,	p(-25,-20),p(  0,-20),p(25,10),p(10,10),
		$(l_o_half,$(l_c_forward,$(r_c_forward,$))))).
$fact(pos( 6, r_w_back,	p(-25, 20),p(  0, 20),p(25,10),p(10,10),
		$(r_o_half,$(r_c_forward,$(l_c_forward,$))))).
$fact(pos( 7, d_half,	p(-10,  0),p(  5,  0),p(25,20),p(10,5),
		$(l_o_half,$(r_o_half,$(l_c_forward,$(r_c_forward,$)))))).
$fact(pos( 8, l_o_half,	p(-10,-15),p( 15,-15),p(35,20),p(10,5),
		$(l_c_forward,$(goal_l,$(r_c_forward,$))))).
$fact(pos( 9, r_o_half,	p(-10, 15),p( 15, 15),p(35,20),p(10,5),
		$(r_c_forward,$(goal_r,$(l_c_forward,$))))).
$fact(pos(10,l_c_forward,p( -1, -1),p( 30, -5),p(10,20),p(20,7),
		$(goal_l,$(goal,$(goal_r,$(r_c_forward,$)))))).
$fact(pos(11,r_c_forward,p( -1,  1),p( 30,  5),p(10,20),p(20,7),
		$(goal_l,$(goal,$(goal_r,$(r_c_forward,$)))))).
$fact(pos( x, goal,	p( 53,  0),p( 53,  0),p(0,0),p(0,0),
		$)).
$fact(pos( x, goal_l,	p( 53, -3),p( 53, -3),p(0,0),p(0,0),
		$)).
$fact(pos( x, goal_r,	p( 53,  3),p( 53,  3),p(0,0),p(0,0),
		$)).
$fact(pos( x,l_clear,	p(-30,-35),p(-30,-35),p(0,0),p(0,0),
		$)).
$fact(pos( x,r_clear,	p(-30, 35),p(-30, 35),p(0,0),p(0,0),
		$)).

pop(_).

%%======================================================================
%% position table for work
%%

newcell(position_table0).
push(position_table0).

%%    pos(UN, PosName,  InitPos,   HomePos,   Range,   Shift,
%%		PassDestList)

$fact(pos(x1, goalie,	p(-51,  0),p(-50,  0),p(15,15),p( 5,5),
		$(l_w_back,$(r_w_back,$(l_clear,$(r_clear,$)))))).
$fact(pos( 1, goalie,	p(-51,  0),p(-50,  0),p(15,15),p( 5,5),
		$(l_w_back,$(r_w_back,$(l_clear,$(r_clear,$)))))).
$fact(pos( 2, goalie,	p(-51,  0),p(-50,  0),p(15,15),p( 5,5),
		$(l_w_back,$(r_w_back,$(l_clear,$(r_clear,$)))))).
$fact(pos(x2, sweeper,	p(-35,  0),p(-20,  0),p(25,20),p(20,5),
		$(l_w_back,$(r_w_back,$(d_half,$(l_clear,$(r_clear,$))))))).
$fact(pos(x3, l_f_back,	p(-40, -8),p(-30, -8),p(20,30),p(20,5),
		$(l_w_back,$(l_o_half,$(l_clear,$))))).
$fact(pos(x4, r_f_back,	p(-40,  8),p(-30,  8),p(20,30),p(20,5),
		$(r_w_back,$(r_o_half,$(r_clear,$))))).
$fact(pos(x5, l_w_back,	p(-25,-20),p( -0,-20),p(25,10),p(10,10),
		$(l_o_half,$(l_c_forward,$(r_c_forward,$))))).
$fact(pos(x6, r_w_back,	p(-25, 20),p( -0, 20),p(25,10),p(10,10),
		$(r_o_half,$(r_c_forward,$(l_c_forward,$))))).
$fact(pos(x7, d_half,	p(-10,  0),p(  5,  0),p(25,20),p(10,5),
		$(l_o_half,$(r_o_half,$(l_c_forward,$(r_c_forward,$)))))).
$fact(pos(x8, l_o_half,	p(-10,-15),p( 15,-15),p(35,20),p(10,5),
		$(l_c_forward,$(goal_l,$(r_c_forward,$))))).
$fact(pos(x9, r_o_half,	p(-10, 15),p( 15, 15),p(35,20),p(10,5),
		$(r_c_forward,$(goal_r,$(l_c_forward,$))))).
$fact(pos(x10,l_c_forward,p( -1, -1),p( 30, -5),p(10,20),p(20,5),
		$(goal_l,$(goal,$(goal_r,$(r_c_forward,$)))))).
$fact(pos( 3,l_c_forward,p( -1, -1),p( 30, -5),p(10,20),p(20,5),
		$(goal_l,$(goal,$(goal_r,$(r_c_forward,$)))))).
$fact(pos(x11,r_c_forward,p( -1,  1),p( 30,  5),p(10,20),p(20,5),
		$(goal_l,$(goal,$(goal_r,$(r_c_forward,$)))))).
$fact(pos( 4,r_c_forward,p( -1,  1),p( 30,  5),p(10,20),p(20,5),
		$(goal_l,$(goal,$(goal_r,$(r_c_forward,$)))))).
$fact(pos( x, goal,	p( 53,  0),p( 53,  0),p(0,0),p(0,0),
		$)).
$fact(pos( x, goal_l,	p( 53, -3),p( 53, -3),p(0,0),p(0,0),
		$)).
$fact(pos( x, goal_r,	p( 53,  3),p( 53,  3),p(0,0),p(0,0),
		$)).
$fact(pos( x,l_clear,	p(-30,-35),p(-30,-35),p(0,0),p(0,0),
		$)).
$fact(pos( x,r_clear,	p(-30, 35),p(-30, 35),p(0,0),p(0,0),
		$)).

pop(_).

%%======================================================================
%% push work
%%

newcell(work).
push(work).

