// -*- Mode:C++ -*-
//Header:
//File: unify.h
//Author: NODA Itsuki
//Date: 1998/3/27
//EndHeader:

//ModifyHistory:
// 1998/3/27: Create this file
//EndModifyHistory:

//Memo:
//EndMemo:

#ifndef _unify_h_
#define _unify_h_
////////////////////////////////////////////////////////////////////////
#include <stdlib.h>
#include <iostream.h>
#include "mmtoolT.h"
#include "gdata.h"

//======================================================================
// bind
//

class UnifyInfoHeap ;
class Binding ;

class InstGData {
   public: 
      GData data ;
      Binding* binding ;

      inline InstGData(const GData d = NullSymbol, 
		       const Binding* b = NULLPTR) 
      {
	 data = d ; binding = (Binding*)b ;
      } ;

      inline void init() {
	 data.init() ;	
	 binding = NULLPTR ;
      } ;

      inline InstGData operator=(const InstGData& src) {
	 data = src.data ;
	 binding = src.binding ;
	 return *this ;
      } ;
     
      friend ostream& operator<< (ostream& o, const InstGData& dt) ;
} ; 

class Bind {
   public:
      GData variable ;
      InstGData value ;

      inline Bind(GData var, InstGData val) {
	 variable = var ; 
	 value = val ;
      } ;
      inline Bind() {} ;
     
      inline void init() { 
	 variable.init() ;
	 value.init() ;
      } ;

      friend ostream& operator<< (ostream& o, const Bind& bind) ;

} ;

//======================================================================
// newvartable
//
class NewVarEntry {
   public:
      InstGData variable ;
      GData newvariable ;

      inline NewVarEntry() {} ;
      inline NewVarEntry(InstGData var, GData newvar) {
	 variable = var ;
	 newvariable = newvar ;
      } ;
      inline NewVarEntry(GData var, Binding* binding, GData newvar) {
	 NewVarEntry(InstGData(var,binding),newvar) ;
      } ;

      inline void init() {
	 variable.init() ;
	 newvariable.init() ;
      } ;

      friend ostream& operator<< (ostream& o, const NewVarEntry& nventry) ;

} ;

typedef SList<NewVarEntry> NewVarList ;

extern GData newVariable(const InstGData& var, 
			 NewVarList* nvlist,
			 UnifyInfoHeap* uiheap) ;


		    

//======================================================================
// binding
//
typedef SList<Bind> BindList ;

class Binding {
   public:
      BindList* bindlist ;

      inline Binding() { init() ; } ;

      inline void init() { 
	 bindlist = NULLPTR ; 
      } ;

      Bind* findBind(const GData var) ;
      InstGData findBindValue(const GData var) ;

      BindList* pushBindList(GData var, InstGData val, UnifyInfoHeap* uiheap) ;

      GData newVariable(GData var, UnifyInfoHeap* uiheap) ;

      friend ostream& operator<< (ostream& o, const Binding& binding) ;
      friend ostream& operator<< (ostream& o, const Binding* binding) ;
} ;

//======================================================================
// Unify Info Heap
//
typedef SListHeap<Bind> BindListHeap ;
typedef AHeap<Binding> BindingHeap ;
typedef SListHeap<NewVarEntry> NewVarListHeap ;

class UnifyInfoHeap {
   public:
      BindListHeap bindListHeap ;
      BindingHeap bindingHeap ;
      NewVarListHeap newvarListHeap ;

      inline void init(UInt size = 0, bool verbosep = false) {
	 if(size>0) {
	    bindListHeap.expand(size,verbosep) ;
	    bindingHeap.expand(size,verbosep) ;
	    newvarListHeap.expand(size,verbosep) ;
	 }
      } ;

      inline void reset() {
	 bindListHeap.reset() ;
	 bindingHeap.reset() ;
	 newvarListHeap.reset() ;
      } ;

      inline BindList* pushBindList(Bind bind, BindList* blist) {
	 return bindListHeap.push(bind,blist) ;
      } ;

      inline Binding* newBinding() {
	 return bindingHeap.get() ;
      } ;

      inline NewVarList* pushNewVarList(NewVarEntry nventry, 
					NewVarList* nvlist) 
      {
	 return this->newvarListHeap.push(nventry,nvlist) ;
      } ;

      class rewindPointer {
	 public:
	    UInt bindListHeapPointer ;
	    UInt bindingHeapPointer ;

	    inline rewindPointer() {
	       bindListHeapPointer = 0 ;
	       bindingHeapPointer = 0 ;
	    } ;
	    inline rewindPointer(const rewindPointer& up) {
	       bindListHeapPointer = up.bindListHeapPointer ;
	       bindingHeapPointer = up.bindingHeapPointer ;
	    } ;
	    inline rewindPointer(const UInt blisthp, const UInt bindinghp) {
	       bindListHeapPointer = blisthp ;
	       bindingHeapPointer = bindinghp ;
	    } ;
	    
	    inline rewindPointer operator= (const rewindPointer& up) {
	       bindListHeapPointer = up.bindListHeapPointer ;
	       bindingHeapPointer = up.bindingHeapPointer ;
	       return *this ;
	    } ;

      } ;

      inline rewindPointer rewindPoint() {
	 return rewindPointer(bindListHeap.toppos(),bindingHeap.toppos()) ;
      } ;
      inline bool rewind(const rewindPointer& rp) {
	 bindListHeap.rewind(rp.bindListHeapPointer) ;
	 bindingHeap.rewind(rp.bindingHeapPointer) ;
	 return true ;
      } ;

      inline void describe(ostream& o, bool detailp = true) const {
	 o << "#UIHeap[" ;
	 o << bindListHeap ;
//	 o << "," << bindingHeap ;
	 o << "," << newvarListHeap ;
	 o << "]" ;
      } ;


      inline friend ostream& operator<<(ostream& o, 
					const UnifyInfoHeap& uiheap) 
      {
	 uiheap.describe(o,false) ;
	 return o ;
      } ;
} ;	
     
//======================================================================
// Binding State
//
class BindingState {
   public:
      Binding* binding ;
      UnifyInfoHeap* uiheap ;
      BindList* bindlist ;
      UInt bindListHeapPtr ;
      UInt bindingHeapPtr ;

      inline BindingState() { 
	 binding = NULLPTR ;
	 uiheap = NULLPTR ;
	 bindlist = NULLPTR ; 
	 bindListHeapPtr = 0 ; 
	 bindingHeapPtr = 0 ;
      } ;
      inline BindingState(Binding* b, UnifyInfoHeap* uih) {
	 memory(b, uih) ;
      } ;
      inline BindingState(const BindingState& src) {
	 binding = src.binding ;
	 uiheap = src.uiheap ;
	 bindlist = src.bindlist ;
	 bindListHeapPtr = src.bindListHeapPtr ;
	 bindingHeapPtr = src.bindingHeapPtr ;
      } ;

      inline BindingState operator=(const BindingState& src) {
	 binding = src.binding ;
	 uiheap = src.uiheap ;
	 bindlist = src.bindlist ;
	 bindListHeapPtr = src.bindListHeapPtr ;
	 bindingHeapPtr = src.bindingHeapPtr ;
	 return *this ;
      } ;

      inline bool memory(Binding* b, UnifyInfoHeap* uih) {
	 binding = b ;
	 uiheap = uih ;
	 bindlist = binding->bindlist ;
	 bindListHeapPtr = uiheap->bindListHeap.toppos() ;
	 bindingHeapPtr = uiheap->bindingHeap.toppos() ;
	 return true ;
      } ;

      inline bool rewind() {
	 binding->bindlist = bindlist ;
	 uiheap->bindListHeap.rewind(bindListHeapPtr) ;
	 uiheap->bindingHeap.rewind(bindingHeapPtr) ;
      } ;

      inline void describe(ostream& o, bool detailp = true) const {
	 o << "#BindingState[" << bindlist << ":" << bindListHeapPtr 
	   << "," << bindingHeapPtr << "]" ;
	 if(detailp) {
	    o << endl << "   BindList=" << *bindlist << endl;
	 }
      } ;

      inline friend ostream& operator<< (ostream& o, const BindingState& bs) {
	 bs.describe(o,false) ;
	 return o ;
      } ;
	    
} ;

inline BindingState memoryBindingState(Binding* binding, 
				       UnifyInfoHeap* uiheap) 
{
   return BindingState(binding,uiheap) ;
} ;

inline bool memoryBindingState(BindingState& bstate, 
			       Binding* binding,
			       UnifyInfoHeap* uiheap) 
{
   return bstate.memory(binding,uiheap) ;
} ;


inline bool rewindBindingState(BindingState& bstate) 
{
   return bstate.rewind() ;
} ;

inline bool rewindBindingState(BindingState* bstate) 
{
   return bstate->rewind() ;
} ;
			       

     

//==================================================
// Binding State 2 
//

class BindingState2 {
   public:
      Binding* xbinding ;
      Binding* ybinding ;
      UnifyInfoHeap* uiheap ;
      BindList* xbindlist ;
      BindList* ybindlist ;
      UInt bindListHeapPtr ;
      UInt bindingHeapPtr ;

      inline BindingState2() { 
	 xbinding = NULLPTR ;
	 ybinding = NULLPTR ;
	 uiheap = NULLPTR ;
	 xbindlist = NULLPTR ; 
	 ybindlist = NULLPTR ; 
	 bindListHeapPtr = 0 ; 
	 bindingHeapPtr = 0 ;
      } ;
      inline BindingState2(Binding* xb, Binding* yb, UnifyInfoHeap* uih) {
	 memory(xb, yb, uih) ;
      } ;
      inline BindingState2(const BindingState2& src) {
	 xbinding = src.xbinding ;
	 ybinding = src.ybinding ;
	 uiheap = src.uiheap ;
	 xbindlist = src.xbindlist ;
	 ybindlist = src.ybindlist ;
	 bindListHeapPtr = src.bindListHeapPtr ;
	 bindingHeapPtr = src.bindingHeapPtr ;
      } ;

      inline BindingState2 operator=(const BindingState2& src) {
	 xbinding = src.xbinding ;
	 ybinding = src.ybinding ;
	 uiheap = src.uiheap ;
	 xbindlist = src.xbindlist ;
	 ybindlist = src.ybindlist ;
	 bindListHeapPtr = src.bindListHeapPtr ;
	 bindingHeapPtr = src.bindingHeapPtr ;
	 return *this ;
      } ;

      inline bool memory(Binding* xb, Binding* yb, UnifyInfoHeap* uih) {
	 xbinding = xb ;
	 ybinding = yb ;
	 uiheap = uih ;
	 xbindlist = ybinding->bindlist ;
	 ybindlist = ybinding->bindlist ;
	 bindListHeapPtr = uiheap->bindListHeap.toppos() ;
	 bindingHeapPtr = uiheap->bindingHeap.toppos() ;
	 return true ;
      } ;

      inline bool rewind() {
	 xbinding->bindlist = xbindlist ;
	 ybinding->bindlist = ybindlist ;
	 uiheap->bindListHeap.rewind(bindListHeapPtr) ;
	 uiheap->bindingHeap.rewind(bindingHeapPtr) ;
      } ;

      inline void describe(ostream& o, bool detailp = true) const {
	 o << "#BindingState2[" << xbindlist << ":" << ybindlist << ":" 
	   << bindListHeapPtr << "," << bindingHeapPtr << "]" ;
	 if(detailp) {
	    o << endl << "   XBindList=" << *xbindlist << endl;
	    o << endl << "   YBindList=" << *ybindlist << endl;
	 }
      } ;

      inline friend ostream& operator<< (ostream& o, const BindingState2& bs) {
	 bs.describe(o,false) ;
	 return o ;
      } ;
	    
} ;

inline BindingState2 memoryBindingState2(Binding* xbinding, 
					 Binding* ybinding, 
					 UnifyInfoHeap* uiheap) 
{
   return BindingState2(xbinding,ybinding,uiheap) ;
} ;

inline bool memoryBindingState2(BindingState2& bstate, 
				Binding* xbinding,
				Binding* ybinding,
				UnifyInfoHeap* uiheap) 
{
   return bstate.memory(xbinding,ybinding,uiheap) ;
} ;


inline bool rewindBindingState2(BindingState2& bstate) 
{
   return bstate.rewind() ;
} ;

inline bool rewindBindingState2(BindingState2* bstate) 
{
   return bstate->rewind() ;
} ;
			       

     

//======================================================================
// isAssigned
//

extern Bind* isAssigned(GData var, Binding* binding) ;
inline Bind* isAssigned(InstGData dt) {
   return isAssigned(dt.data, dt.binding) ; 
};

extern InstGData isAssignedIGD(GData var, Binding* binding) ;
inline InstGData isAssignedIGD(InstGData dt) {
   return isAssignedIGD(dt.data, dt.binding) ; 
};

extern bool isGround(GData var, Binding* binding) ;
inline bool isGround(InstGData dt) {
   return isGround(dt.data, dt.binding) ;
};

//======================================================================
// unify
//

extern bool Unify(InstGData x, InstGData y, UnifyInfoHeap* uiheap) ;
extern bool unify(InstGData x, InstGData y, UnifyInfoHeap* uiheap) ;
extern bool unifyGTerm(InstGData x, InstGData y, UnifyInfoHeap* uiheap) ;
extern void linkBind(InstGData var, InstGData val, UnifyInfoHeap* uiheap) ;

//======================================================================
// fetch
//

extern InstGData fetch(InstGData idt) ;
inline InstGData fetch(GData dt, Binding* binding) {
   return fetch(InstGData(dt,binding)) ;
} 

//======================================================================
// fetch-value
//

extern GData fetchValue(const InstGData& idt, UnifyInfoHeap* uiheap) ;
extern GData fetchValueBody(const InstGData& idt, UnifyInfoHeap* uiheap,
			    NewVarList* nvlist) ;
extern GData fetchValueGTerm(const InstGData& idt, UnifyInfoHeap* uiheap, 
			     NewVarList* nvlist) ;






	  





////////////////////////////////////////////////////////////////////////
#endif
